@@ -1,9 +1,9 @@
-/*  Lunzip - Decompressor for lzip files
-    Copyright (C) 2010, 2011, 2012, 2013, 2014 Antonio Diaz Diaz.
+/*  Lunzip - Decompressor for the lzip format
+    Copyright (C) 2010-2017 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -23,6 +23,7 @@
 
 #define _FILE_OFFSET_BITS 64
 
+#include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <limits.h>
@@ -62,57 +63,62 @@
 #error "Environments where CHAR_BIT != 8 are not supported."
 #endif
 
+int verbosity = 0;
 
 const char * const Program_name = "Lunzip";
 const char * const program_name = "lunzip";
-const char * const program_year = "2014";
+const char * const program_year = "2017";
 const char * invocation_name = 0;
 
-struct { const char * from; const char * to; } const known_extensions[] = {
+const struct { const char * from; const char * to; } known_extensions[] = {
   { ".lz",  ""     },
   { ".tlz", ".tar" },
   { 0,      0      } };
 
+enum Mode { m_compress, m_decompress, m_list, m_test };
+
 char * output_filename = 0;
 int outfd = -1;
-int verbosity = 0;
-const mode_t usr_rw = S_IRUSR | S_IWUSR;
-const mode_t all_rw = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
-mode_t outfd_mode = S_IRUSR | S_IWUSR;
 bool delete_output_on_interrupt = false;
 
 
 static void show_help( void )
   {
-  printf( "Lunzip is a decompressor for lzip files. It is written in C and its\n"
+  printf( "Lunzip is a decompressor for the lzip format. It is written in C and its\n"
           "small size makes it well suited for embedded devices or software\n"
-          "installers that need to decompress files but do not need compression\n"
+          "installers that need to decompress files but don't need compression\n"
           "capabilities. Lunzip is fully compatible with lzip-1.4 or newer.\n"
           "\nLunzip provides a 'low memory' mode able to decompress any file using as\n"
           "little memory as 50 kB, irrespective of the dictionary size used to\n"
           "compress the file. To activate it, specify the size of the output buffer\n"
           "with the '--buffer-size' option and lunzip will use the decompressed\n"
           "file as dictionary for distances beyond the buffer size. Of course, the\n"
-          "smaller the output buffer size used in relation to the dictionary size,\n"
-          "the more accesses to disk are needed and the slower the decompression is.\n"
-          "This 'low memory' mode only works when decompressing to a regular file.\n" );
-  printf( "\nUsage: %s [options] [files]\n", invocation_name );
+          "smaller the buffer size used in relation to the dictionary size, the\n"
+          "more accesses to disk are needed and the slower the decompression is.\n"
+          "This 'low memory' mode only works when decompressing to a regular file\n"
+          "and is intended for systems without enough memory (RAM + swap) to keep\n"
+          "the whole dictionary at once.\n"
+          "\nUsage: %s [options] [files]\n", invocation_name );
   printf( "\nOptions:\n"
           "  -h, --help                 display this help and exit\n"
           "  -V, --version              output version information and exit\n"
-          "  -c, --stdout               send output to standard output\n"
+          "  -a, --trailing-error       exit with error status if trailing data\n"
+          "  -c, --stdout               write to standard output, keep input files\n"
           "  -d, --decompress           decompress (this is the default)\n"
           "  -f, --force                overwrite existing output files\n"
           "  -k, --keep                 keep (don't delete) input files\n"
-          "  -o, --output=<file>        if reading stdin, place the output into <file>\n"
+          "  -l, --list                 print (un)compressed file sizes\n"
+          "  -o, --output=<file>        if reading standard input, write to <file>\n"
           "  -q, --quiet                suppress all messages\n"
           "  -t, --test                 test compressed file integrity\n"
           "  -u, --buffer-size=<bytes>  set output buffer size in bytes\n"
           "  -v, --verbose              be verbose (a 2nd -v gives more)\n"
-          "If no file names are given, lunzip decompresses from standard input to\n"
-          "standard output.\n"
+          "If no file names are given, or if a file is '-', lunzip decompresses\n"
+          "from standard input to standard output.\n"
           "Numbers may be followed by a multiplier: k = kB = 10^3 = 1000,\n"
           "Ki = KiB = 2^10 = 1024, M = 10^6, Mi = 2^20, G = 10^9, Gi = 2^30, etc...\n"
+          "Buffer sizes 12 to 29 are interpreted as powers of two, meaning 2^12\n"
+          "to 2^29 bytes.\n"
           "\nExit status: 0 for a normal exit, 1 for environmental problems (file\n"
           "not found, invalid flags, I/O errors, etc), 2 to indicate a corrupt or\n"
           "invalid input file, 3 for an internal consistency error (eg, bug) which\n"
@@ -126,17 +132,27 @@ static void show_version( void )
   {
   printf( "%s %s\n", program_name, PROGVERSION );
   printf( "Copyright (C) %s Antonio Diaz Diaz.\n", program_year );
-  printf( "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n"
+  printf( "License GPLv2+: GNU GPL version 2 or later <http://gnu.org/licenses/gpl.html>\n"
           "This is free software: you are free to change and redistribute it.\n"
           "There is NO WARRANTY, to the extent permitted by law.\n" );
   }
 
 
-static void show_header( const unsigned dictionary_size )
+const char * bad_version( const unsigned version )
   {
+  static char buf[80];
+  snprintf( buf, sizeof buf, "Version %u member format not supported.",
+            version );
+  return buf;
+  }
+
+
+const char * format_ds( const unsigned dictionary_size )
+  {
+  enum { bufsize = 16, factor = 1024 };
+  static char buf[bufsize];
   const char * const prefix[8] =
     { "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi" };
-  enum { factor = 1024 };
   const char * p = "";
   const char * np = "  ";
   unsigned num = dictionary_size, i;
@@ -145,7 +161,15 @@ static void show_header( const unsigned dictionary_size )
   for( i = 0; i < 8 && ( num > 9999 || ( exact && num >= factor ) ); ++i )
     { num /= factor; if( num % factor != 0 ) exact = false;
       p = prefix[i]; np = ""; }
-  fprintf( stderr, "dictionary size %s%4u %sB.  ", np, num, p );
+  snprintf( buf, bufsize, "%s%4u %sB", np, num, p );
+  return buf;
+  }
+
+
+static void show_header( const unsigned dictionary_size )
+  {
+  if( verbosity >= 3 )
+    fprintf( stderr, "dictionary %s.  ", format_ds( dictionary_size ) );
   }
 
 
@@ -165,12 +189,11 @@ static unsigned long getnum( const char * const ptr,
 
   if( !errno && tail[0] )
     {
-    int factor = ( tail[1] == 'i' ) ? 1024 : 1000;
-    int exponent = 0, i;
-    bool bad_multiplier = false;
+    const unsigned factor = ( tail[1] == 'i' ) ? 1024 : 1000;
+    int exponent = 0;				/* 0 = bad multiplier */
+    int i;
     switch( tail[0] )
       {
-      case ' ': break;
       case 'Y': exponent = 8; break;
       case 'Z': exponent = 7; break;
       case 'E': exponent = 6; break;
@@ -178,13 +201,10 @@ static unsigned long getnum( const char * const ptr,
       case 'T': exponent = 4; break;
       case 'G': exponent = 3; break;
       case 'M': exponent = 2; break;
-      case 'K': if( factor == 1024 ) exponent = 1; else bad_multiplier = true;
-                break;
-      case 'k': if( factor == 1000 ) exponent = 1; else bad_multiplier = true;
-                break;
-      default : bad_multiplier = true;
+      case 'K': if( factor == 1024 ) exponent = 1; break;
+      case 'k': if( factor == 1000 ) exponent = 1; break;
       }
-    if( bad_multiplier )
+    if( exponent <= 0 )
       {
       show_error( "Bad multiplier in numerical argument.", 0, true );
       exit( 1 );
@@ -208,7 +228,7 @@ static unsigned long getnum( const char * const ptr,
 static int get_dict_size( const char * const arg )
   {
   char * tail;
-  int bits = strtol( arg, &tail, 0 );
+  const long bits = strtol( arg, &tail, 0 );
   if( bits >= min_dictionary_bits &&
       bits <= max_dictionary_bits && *tail == 0 )
     return ( 1 << bits );
@@ -216,36 +236,44 @@ static int get_dict_size( const char * const arg )
   }
 
 
+void set_mode( enum Mode * const program_modep, const enum Mode new_mode )
+  {
+  if( *program_modep != m_compress && *program_modep != new_mode )
+    {
+    show_error( "Only one operation can be specified.", 0, true );
+    exit( 1 );
+    }
+  *program_modep = new_mode;
+  }
+
+
 static int extension_index( const char * const name )
   {
-  int i;
-  for( i = 0; known_extensions[i].from; ++i )
+  int eindex;
+  for( eindex = 0; known_extensions[eindex].from; ++eindex )
     {
-    const char * const ext = known_extensions[i].from;
-    if( strlen( name ) > strlen( ext ) &&
-        strncmp( name + strlen( name ) - strlen( ext ), ext, strlen( ext ) ) == 0 )
-      return i;
+    const char * const ext = known_extensions[eindex].from;
+    const unsigned name_len = strlen( name );
+    const unsigned ext_len = strlen( ext );
+    if( name_len > ext_len &&
+        strncmp( name + name_len - ext_len, ext, ext_len ) == 0 )
+      return eindex;
     }
   return -1;
   }
 
 
-static int open_instream( const char * const name, struct stat * const in_statsp,
-                          const bool no_ofile )
+int open_instream( const char * const name, struct stat * const in_statsp,
+                   const bool no_ofile, const bool reg_only )
   {
-  int infd;
-  infd = open( name, O_RDONLY | O_BINARY );
+  int infd = open( name, O_RDONLY | O_BINARY );
   if( infd < 0 )
-    {
-    if( verbosity >= 0 )
-      fprintf( stderr, "%s: Can't open input file '%s': %s.\n",
-               program_name, name, strerror( errno ) );
-    }
+    show_file_error( name, "Can't open input file", errno );
   else
     {
     const int i = fstat( infd, in_statsp );
     const mode_t mode = in_statsp->st_mode;
-    const bool can_read = ( i == 0 &&
+    const bool can_read = ( i == 0 && !reg_only &&
                             ( S_ISBLK( mode ) || S_ISCHR( mode ) ||
                               S_ISFIFO( mode ) || S_ISSOCK( mode ) ) );
     if( i != 0 || ( !S_ISREG( mode ) && ( !can_read || !no_ofile ) ) )
@@ -254,7 +282,7 @@ static int open_instream( const char * const name, struct stat * const in_statsp
         fprintf( stderr, "%s: Input file '%s' is not a regular file%s.\n",
                  program_name, name,
                  ( can_read && !no_ofile ) ?
-                 " and '--stdout' was not specified" : "" );
+                 ",\n        and '--stdout' was not specified" : "" );
       close( infd );
       infd = -1;
       }
@@ -264,7 +292,7 @@ static int open_instream( const char * const name, struct stat * const in_statsp
 
 
 /* assure at least a minimum size for buffer 'buf' */
-static void * resize_buffer( void * buf, const int min_size )
+void * resize_buffer( void * buf, const unsigned min_size )
   {
   if( buf ) buf = realloc( buf, min_size );
   else buf = malloc( min_size );
@@ -277,43 +305,48 @@ static void * resize_buffer( void * buf, const int min_size )
   }
 
 
-static void set_d_outname( const char * const name, const int i )
+static void set_d_outname( const char * const name, const int eindex )
   {
-  if( i >= 0 )
+  const unsigned name_len = strlen( name );
+  if( eindex >= 0 )
     {
-    const char * const from = known_extensions[i].from;
-    if( strlen( name ) > strlen( from ) )
+    const char * const from = known_extensions[eindex].from;
+    const unsigned from_len = strlen( from );
+    if( name_len > from_len )
       {
-      output_filename = resize_buffer( output_filename, strlen( name ) +
-                                       strlen( known_extensions[0].to ) + 1 );
+      output_filename = resize_buffer( output_filename, name_len +
+                                       strlen( known_extensions[eindex].to ) + 1 );
       strcpy( output_filename, name );
-      strcpy( output_filename + strlen( name ) - strlen( from ),
-              known_extensions[i].to );
+      strcpy( output_filename + name_len - from_len, known_extensions[eindex].to );
       return;
       }
     }
-  output_filename = resize_buffer( output_filename, strlen( name ) + 4 + 1 );
+  output_filename = resize_buffer( output_filename, name_len + 4 + 1 );
   strcpy( output_filename, name );
   strcat( output_filename, ".out" );
   if( verbosity >= 1 )
-    fprintf( stderr, "%s: Can't guess original name for '%s' -- using '%s'.\n",
+    fprintf( stderr, "%s: Can't guess original name for '%s' -- using '%s'\n",
              program_name, name, output_filename );
   }
 
 
-static bool open_outstream( const bool force )
+static bool open_outstream( const bool force, const bool from_stdin )
   {
+  const mode_t usr_rw = S_IRUSR | S_IWUSR;
+  const mode_t all_rw = usr_rw | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
+  const mode_t outfd_mode = from_stdin ? all_rw : usr_rw;
   int flags = O_APPEND | O_CREAT | O_RDWR | O_BINARY;
   if( force ) flags |= O_TRUNC; else flags |= O_EXCL;
 
   outfd = open( output_filename, flags, outfd_mode );
-  if( outfd < 0 && verbosity >= 0 )
+  if( outfd >= 0 ) delete_output_on_interrupt = true;
+  else if( verbosity >= 0 )
     {
     if( errno == EEXIST )
       fprintf( stderr, "%s: Output file '%s' already exists, skipping.\n",
                program_name, output_filename );
     else
-      fprintf( stderr, "%s: Can't create output file '%s': %s.\n",
+      fprintf( stderr, "%s: Can't create output file '%s': %s\n",
                program_name, output_filename, strerror( errno ) );
     }
   return ( outfd >= 0 );
@@ -351,7 +384,11 @@ static void close_and_set_permissions( const struct stat * const in_statsp )
           fchmod( outfd, mode & ~( S_ISUID | S_ISGID | S_ISVTX ) ) != 0 )
         warning = true;
     }
-  if( close( outfd ) != 0 ) cleanup_and_fail( 1 );
+  if( close( outfd ) != 0 )
+    {
+    show_error( "Error closing output file", errno, false );
+    cleanup_and_fail( 1 );
+    }
   outfd = -1;
   delete_output_on_interrupt = false;
   if( in_statsp )
@@ -366,8 +403,45 @@ static void close_and_set_permissions( const struct stat * const in_statsp )
   }
 
 
-static int decompress( const int buffer_size, const int infd,
-                       struct Pretty_print * const pp, const bool testing )
+static unsigned char xdigit( const unsigned value )
+  {
+  if( value <= 9 ) return '0' + value;
+  if( value <= 15 ) return 'A' + value - 10;
+  return 0;
+  }
+
+
+static bool show_trailing_data( const uint8_t * const data, const int size,
+                                struct Pretty_print * const pp, const bool all,
+                                const bool ignore_trailing )
+  {
+  if( verbosity >= 4 || !ignore_trailing )
+    {
+    int i;
+    char buf[80];
+    unsigned len = max( 0, snprintf( buf, sizeof buf, "%strailing data = ",
+                                     all ? "" : "first bytes of " ) );
+    for( i = 0; i < size && len + 2 < sizeof buf; ++i )
+      {
+      buf[len++] = xdigit( data[i] >> 4 );
+      buf[len++] = xdigit( data[i] & 0x0F );
+      buf[len++] = ' ';
+      }
+    if( len < sizeof buf ) buf[len++] = '\'';
+    for( i = 0; i < size && len < sizeof buf; ++i )
+      { if( isprint( data[i] ) ) buf[len++] = data[i]; else buf[len++] = '.'; }
+    if( len < sizeof buf ) buf[len++] = '\'';
+    if( len < sizeof buf ) buf[len] = 0; else buf[sizeof buf - 1] = 0;
+    Pp_show_msg( pp, buf );
+    if( !ignore_trailing ) show_file_error( pp->name, trailing_msg, 0 );
+    }
+  return ignore_trailing;
+  }
+
+
+static int decompress( const int infd, struct Pretty_print * const pp,
+                       const unsigned buffer_size,
+                       const bool ignore_trailing, const bool testing )
   {
   unsigned long long partial_file_pos = 0;
   struct Range_decoder rdec;
@@ -381,47 +455,46 @@ static int decompress( const int buffer_size, const int infd,
 
   for( first_member = true; ; first_member = false )
     {
-    int result;
+    int result, size;
     unsigned dictionary_size;
     File_header header;
     struct LZ_decoder decoder;
     Rd_reset_member_position( &rdec );
-    Rd_read_data( &rdec, header, Fh_size );
+    size = Rd_read_data( &rdec, header, Fh_size );
     if( Rd_finished( &rdec ) )			/* End Of File */
       {
-      if( first_member )
+      if( first_member || Fh_verify_prefix( header, size ) )
         { Pp_show_msg( pp, "File ends unexpectedly at member header." );
           retval = 2; }
+      else if( size > 0 && !show_trailing_data( header, size, pp,
+                                                true, ignore_trailing ) )
+        retval = 2;
       break;
       }
     if( !Fh_verify_magic( header ) )
       {
-      if( !first_member ) break;		/* trailing garbage */
-      Pp_show_msg( pp, "Bad magic number (file not in lzip format)." );
-      retval = 2; break;
+      if( first_member )
+        { show_file_error( pp->name, bad_magic_msg, 0 ); retval = 2; }
+      else if( !show_trailing_data( header, size, pp, false, ignore_trailing ) )
+        retval = 2;
+      break;
       }
     if( !Fh_verify_version( header ) )
       {
-      if( verbosity >= 0 )
-        { Pp_show_msg( pp, 0 );
-          fprintf( stderr, "Version %d member format not supported.\n",
-                   Fh_version( header ) ); }
+      Pp_show_msg( pp, bad_version( Fh_version( header ) ) );
       retval = 2; break;
       }
     dictionary_size = Fh_get_dictionary_size( header );
-    if( dictionary_size < min_dictionary_size ||
-        dictionary_size > max_dictionary_size )
-      { Pp_show_msg( pp, "Invalid dictionary size in member header." );
-        retval = 2; break; }
+    if( !isvalid_ds( dictionary_size ) )
+      { Pp_show_msg( pp, bad_dict_msg ); retval = 2; break; }
 
     if( verbosity >= 2 || ( verbosity == 1 && first_member ) )
-      { Pp_show_msg( pp, 0 );
-        if( verbosity >= 3 ) show_header( dictionary_size ); }
+      { Pp_show_msg( pp, 0 ); show_header( dictionary_size ); }
 
     if( !LZd_init( &decoder, &rdec, buffer_size, dictionary_size, outfd ) )
       {
-      show_error( "Not enough memory. Try a smaller output buffer size.", 0, false );
-      cleanup_and_fail( 1 );
+      Pp_show_msg( pp, "Not enough memory. Try a smaller output buffer size." );
+      retval = 1; break;
       }
     result = LZd_decode_member( &decoder, pp );
     partial_file_pos += Rd_member_position( &rdec );
@@ -431,20 +504,17 @@ static int decompress( const int buffer_size, const int infd,
       if( verbosity >= 0 && result <= 2 )
         {
         Pp_show_msg( pp, 0 );
-        if( result == 2 )
-          fprintf( stderr, "File ends unexpectedly at pos %llu.\n",
-                   partial_file_pos );
-        else
-          fprintf( stderr, "Decoder error at pos %llu.\n", partial_file_pos );
+        fprintf( stderr, "%s at pos %llu\n", ( result == 2 ) ?
+                 "File ends unexpectedly" : "Decoder error", partial_file_pos );
         }
       retval = 2; break;
       }
     if( verbosity >= 2 )
-      { fprintf( stderr, testing ? "ok\n" : "done\n" ); Pp_reset( pp ); }
+      { fputs( testing ? "ok\n" : "done\n", stderr ); Pp_reset( pp ); }
     }
   Rd_free( &rdec );
   if( verbosity == 1 && retval == 0 )
-    fprintf( stderr, testing ? "ok\n" : "done\n" );
+    fputs( testing ? "ok\n" : "done\n", stderr );
   return retval;
   }
 
@@ -467,18 +537,26 @@ static void set_signals( void )
 
 void show_error( const char * const msg, const int errcode, const bool help )
   {
-  if( verbosity >= 0 )
+  if( verbosity < 0 ) return;
+  if( msg && msg[0] )
     {
-    if( msg && msg[0] )
-      {
-      fprintf( stderr, "%s: %s", program_name, msg );
-      if( errcode > 0 ) fprintf( stderr, ": %s.", strerror( errcode ) );
-      fprintf( stderr, "\n" );
-      }
-    if( help )
-      fprintf( stderr, "Try '%s --help' for more information.\n",
-               invocation_name );
+    fprintf( stderr, "%s: %s", program_name, msg );
+    if( errcode > 0 ) fprintf( stderr, ": %s", strerror( errcode ) );
+    fputc( '\n', stderr );
     }
+  if( help )
+    fprintf( stderr, "Try '%s --help' for more information.\n",
+             invocation_name );
+  }
+
+
+void show_file_error( const char * const filename, const char * const msg,
+                      const int errcode )
+  {
+  if( verbosity < 0 ) return;
+  fprintf( stderr, "%s: %s: %s", program_name, filename, msg );
+  if( errcode > 0 ) fprintf( stderr, ": %s", strerror( errcode ) );
+  fputc( '\n', stderr );
   }
 
 
@@ -492,29 +570,32 @@ void internal_error( const char * const msg )
 
 int main( const int argc, const char * const argv[] )
   {
-  const char * input_filename = "";
   const char * default_output_filename = "";
   const char ** filenames = 0;
   int num_filenames = 0;
-  int buffer_size = max_dictionary_size;
+  unsigned buffer_size = max_dictionary_size;
   int infd = -1;
+  enum Mode program_mode = m_compress;
   int argind = 0;
   int retval = 0;
   int i;
   bool filenames_given = false;
   bool force = false;
+  bool ignore_trailing = true;
   bool keep_input_files = false;
-  bool testing = false;
+  bool stdin_used = false;
   bool to_stdout = false;
   struct Pretty_print pp;
 
   const struct ap_Option options[] =
     {
+    { 'a', "trailing-error",  ap_no  },
     { 'c', "stdout",          ap_no  },
     { 'd', "decompress",      ap_no  },
     { 'f', "force",           ap_no  },
     { 'h', "help",            ap_no  },
     { 'k', "keep",            ap_no  },
+    { 'l', "list",            ap_no  },
     { 'n', "threads",         ap_yes },
     { 'o', "output",          ap_yes },
     { 'q', "quiet",           ap_no  },
@@ -541,15 +622,17 @@ int main( const int argc, const char * const argv[] )
     if( !code ) break;					/* no more options */
     switch( code )
       {
+      case 'a': ignore_trailing = false; break;
       case 'c': to_stdout = true; break;
-      case 'd': testing = false; break;
+      case 'd': set_mode( &program_mode, m_decompress ); break;
       case 'f': force = true; break;
       case 'h': show_help(); return 0;
       case 'k': keep_input_files = true; break;
+      case 'l': set_mode( &program_mode, m_list ); break;
       case 'n': break;
       case 'o': default_output_filename = arg; break;
       case 'q': verbosity = -1; break;
-      case 't': testing = true; break;
+      case 't': set_mode( &program_mode, m_test ); break;
       case 'u': buffer_size = get_dict_size( arg ); break;
       case 'v': if( verbosity < 4 ) ++verbosity; break;
       case 'V': show_version(); return 0;
@@ -562,9 +645,6 @@ int main( const int argc, const char * const argv[] )
   setmode( STDOUT_FILENO, O_BINARY );
 #endif
 
-  if( testing )
-    outfd = -1;
-
   num_filenames = max( 1, ap_arguments( &parser ) - argind );
   filenames = resize_buffer( filenames, num_filenames * sizeof filenames[0] );
   filenames[0] = "-";
@@ -575,28 +655,38 @@ int main( const int argc, const char * const argv[] )
     if( strcmp( filenames[i], "-" ) != 0 ) filenames_given = true;
     }
 
+  if( program_mode == m_list )
+    return list_files( filenames, num_filenames, ignore_trailing );
+
+  if( program_mode == m_test )
+    outfd = -1;
+  else if( program_mode == m_compress )
+    program_mode = m_decompress;		/* default mode */
+
   if( buffer_size < max_dictionary_size )
     {
-    if( to_stdout || testing )
+    bool from_stdin = false;
+    if( to_stdout || program_mode == m_test )
       { show_error( "'--buffer-size' is incompatible with '--stdout' and '--test'.", 0, false );
         return 1; }
-    if( !default_output_filename[0] )
-      for( i = 0; i < num_filenames; ++i )
-        if( !filenames[i][0] || strcmp( filenames[i], "-" ) == 0 )
-          { show_error( "Output file must be specified when decompressing from stdin with a\n"
-                        "        reduced buffer size.", 0, false );
-            return 1; }
+    for( i = 0; i < num_filenames; ++i )
+      if( !filenames[i][0] || strcmp( filenames[i], "-" ) == 0 )
+        { from_stdin = true; break; }
+    if( from_stdin && !default_output_filename[0] )
+      { show_error( "Output file must be specified when decompressing from standard input\n"
+                    "        with a reduced buffer size.", 0, false ); return 1; }
     }
 
-  if( !to_stdout && !testing &&
+  if( !to_stdout && program_mode != m_test &&
       ( filenames_given || default_output_filename[0] ) )
     set_signals();
 
-  Pp_init( &pp, filenames, num_filenames );
+  Pp_init( &pp, filenames, num_filenames, verbosity );
 
   output_filename = resize_buffer( output_filename, 1 );
   for( i = 0; i < num_filenames; ++i )
     {
+    const char * input_filename = "";
     int tmp;
     struct stat in_stats;
     const struct stat * in_statsp;
@@ -604,19 +694,18 @@ int main( const int argc, const char * const argv[] )
 
     if( !filenames[i][0] || strcmp( filenames[i], "-" ) == 0 )
       {
-      input_filename = "";
+      if( stdin_used ) continue; else stdin_used = true;
       infd = STDIN_FILENO;
-      if( !testing )
+      if( program_mode != m_test )
         {
         if( to_stdout || !default_output_filename[0] )
           outfd = STDOUT_FILENO;
         else
           {
           output_filename = resize_buffer( output_filename,
-                                           strlen( default_output_filename ) + 1 );
+                              strlen( default_output_filename ) + 1 );
           strcpy( output_filename, default_output_filename );
-          outfd_mode = all_rw;
-          if( !open_outstream( force ) )
+          if( !open_outstream( force, true ) )
             {
             if( retval < 1 ) retval = 1;
             close( infd ); infd = -1;
@@ -628,16 +717,16 @@ int main( const int argc, const char * const argv[] )
     else
       {
       input_filename = filenames[i];
-      infd = open_instream( input_filename, &in_stats, to_stdout || testing );
+      infd = open_instream( input_filename, &in_stats,
+                            to_stdout || program_mode == m_test, false );
       if( infd < 0 ) { if( retval < 1 ) retval = 1; continue; }
-      if( !testing )
+      if( program_mode != m_test )
         {
         if( to_stdout ) outfd = STDOUT_FILENO;
         else
           {
           set_d_outname( input_filename, extension_index( input_filename ) );
-          outfd_mode = usr_rw;
-          if( !open_outstream( force ) )
+          if( !open_outstream( force, false ) )
             {
             if( retval < 1 ) retval = 1;
             close( infd ); infd = -1;
@@ -647,26 +736,41 @@ int main( const int argc, const char * const argv[] )
         }
       }
 
+    Pp_set_name( &pp, input_filename );
     if( isatty( infd ) )
       {
-      show_error( "I won't read compressed data from a terminal.", 0, true );
-      return 1;
+      show_file_error( pp.name,
+                       "I won't read compressed data from a terminal.", 0 );
+      if( retval < 1 ) retval = 1;
+      if( program_mode == m_test ) { close( infd ); infd = -1; continue; }
+      cleanup_and_fail( retval );
+      }
+
+    if( delete_output_on_interrupt && buffer_size < max_dictionary_size )
+      {
+      struct stat st;
+      if( fstat( outfd, &st ) != 0 || !S_ISREG( st.st_mode ) )
+        {
+        if( verbosity >= 0 )
+          fprintf( stderr, "%s: Output file '%s' is not a regular file,\n"
+                   "        and 'low memory' mode has been requested.\n",
+                   program_name, output_filename );
+        if( retval < 1 ) retval = 1;
+        cleanup_and_fail( retval );
+        }
       }
 
-    if( output_filename[0] && !to_stdout && !testing )
-      delete_output_on_interrupt = true;
     in_statsp = input_filename[0] ? &in_stats : 0;
-    Pp_set_name( &pp, input_filename );
-    tmp = decompress( buffer_size, infd, &pp, testing );
+    tmp = decompress( infd, &pp, buffer_size, ignore_trailing, program_mode == m_test );
     if( tmp > retval ) retval = tmp;
-    if( tmp && !testing ) cleanup_and_fail( retval );
+    if( tmp && program_mode != m_test ) cleanup_and_fail( retval );
 
     if( delete_output_on_interrupt )
       close_and_set_permissions( in_statsp );
     if( input_filename[0] )
       {
       close( infd ); infd = -1;
-      if( !keep_input_files && !to_stdout && !testing )
+      if( !keep_input_files && !to_stdout && program_mode != m_test )
         remove( input_filename );
       }
     }
